open Array
open Util
open Data

(* implements my approx bayes factor feature selection *)

module H = IntHashtbl

let bin_search (arr : int array) (target : int) =
  let lin_search (lo : int) (hi : int) =
    let rec lin_search' i = 
      if i >= hi then false
      else if arr.(i) == target then true
      else lin_search' (i+1) in
      lin_search' lo in

  let lo0 = 0 in let hi0 = length arr in let mi0 = (lo0 + hi0) / 2 in
  let rec bin_search' lo mi hi =
    if hi - lo < 5 then lin_search lo hi else
      if arr.(mi) == target then true
      else if arr.(mi) < target then bin_search' mi ((mi+hi)/2) hi
      else bin_search' lo ((lo+mi)/2) mi in
    bin_search' lo0 mi0 hi0

let optimize_single_param baseLik lambda x y wtx g g_y mu0 = (* returns mu, a_g, data probability *)
  let _N = dat_N x in
  let _C = x.datav_numclass in
  let activeH = H.create 5 in
  let baseLikVal = ref one in

    for n = 0 to _N - 1 do
      if bin_search x.datav_xi.(n) g 
      then H.replace activeH n ()
      else baseLikVal := !baseLikVal *@ baseLik.(n);
    done;

  let derivV  = ref 0. in
  let deriv2V = ref 0. in
  let probV   = ref 0. in

  let deriv_both mu =
    derivV  := 0. -. mu *. lambda;
    probV   := 0. -. mu *. mu *. lambda /. 2.;
    let a = ref 0. in let b = ref 0. in
    let s = ref zero in
      H.iter 
        (fun n _ ->
           let v = wtx.(n).(g_y) *@ mu in
             s := zero;
             for y' = 0 to _C - 1 do
               let v = wtx.(n).(y') *@ if y' == g_y then mu else one in
                 s := !s +@ v;
                 if y.(n) == y' then probV := !probV +. v;
             done;
             derivV := !derivV +. (if g_y == y.(n) then 1. else 0.) -. exp (v -. !s);
             probV  := !probV  -. !s;
             a := !a +. exp (v /@ !s) *. exp (v /@ !s);
             b := !b +. exp (v /@ !s);
        ) activeH;
     deriv2V := !a -. !b -. lambda in

  let rec iter mu inum =
    let lastP = !probV in
    let mu'   = mu -. !derivV /. !deriv2V in
      deriv_both mu';
      if inum > 10 || lastP +. 1e-4 > !probV || abs_float (!probV -. lastP) < 1e-6
      then mu', 0. -. !deriv2V, !probV *@ !baseLikVal
      else iter mu' (inum+1) in

    deriv_both mu0;
    iter mu0 1

let compute_likelihood lambda xTr _C priorV wtx y errn current removeit =
  let _N = length wtx in
  let post = ref priorV in
  let dwtx = ref 0. in
  let sum  = ref zero in let maxV = ref zero in
  let toremove = H.create 5 in

    (match current with None -> () | Some (_,_,mu) ->
       post := !post -. (lambda /. 2.) *. mu *. mu);

    H.iter 
      (fun n () ->
         sum := zero; maxV := zero; dwtx := 0.;
         let y_v = 
           match current with 
               None -> -1 
             | Some (y,g,mu) -> (if bin_search xTr.datav_xi.(n) g then dwtx := mu; y) in

         for c = 0 to _C - 1 do
           let v = wtx.(n).(c) +. if c == y_v then !dwtx else 0. in
             sum := addLog !sum v;
             if c == y.(n) then post := !post +. v;
             maxV := max !maxV v;
         done;
         post := !post -. !sum;

         if removeit && y.(n) == y_v then (
           if abs_float (!maxV -. (wtx.(n).(y.(n)) +. !dwtx)) < 1e-10 then
             H.replace toremove n ();
           wtx.(n).(y_v) <- wtx.(n).(y_v) +. !dwtx;
         )
      ) errn;

    if removeit then (
      H.iter (fun n _ -> H.remove errn n) toremove;
      Printf.fprintf stderr "removing %d errors\n" (ih_size toremove);
    );

    !post

let applyActive active x =
  let _N = dat_N x in
  let _C = x.datav_numclass in
  let _F = dat_F x in

  let xi' = init _N (fun _ -> [||]) in

    for n = 0 to _N - 1 do
      let h = H.create 5 in
        for f = 0 to length x.datav_xi.(n) - 1 do
          if H.mem active x.datav_xi.(n).(f) then
            H.replace h (H.find active x.datav_xi.(n).(f)) ();
        done;
      let arr = make (ih_size h) 0 in
      let p   = ref 0 in
        H.iter (fun f _ ->
                  arr.(!p) <- f;
                  incr p) h;
        xi'.(n) <- arr;
    done;
    { datav_bernoulli = x.datav_bernoulli ;
      datav_implicit  = x.datav_implicit  ;
      datav_sampled   = x.datav_sampled   ;
      datav_numclass  = x.datav_numclass  ;
      datav_numfeat   = ih_size active     ;
      datav_indomain  = x.datav_indomain  ;
      datav_weights   = x.datav_weights   ;
      datav_xi        = xi' ;
      datav_xi_c      = [||] ;
      datav_xv        = [||] ;
      datav_xv_c      = [||] }
  
let mem2  h (a,b) = try H.mem (H.find h a) b with Not_found -> false
let find2 h  a b  = H.find (H.find h a) b
let iter2 f h = H.iter (fun a h' -> H.iter (fun b g -> f (a,b) g) h') h

let abffs_selection quiet maxi optimize_l lambda dpp _M xTr yTr xDe yDe xTe yTe sampOption mean_model init_model0 _S =
  let _N = dat_N xTr in
  let _C = xTr.datav_numclass in
  let _F = dat_F xTr in
  let nf = xTr.datav_numfeat in

  for n = 0 to dat_N xTr-1 do fast_sort compare xTr.datav_xi.(n); done;

  let mu = make_ba _F 0. in

  let feature_counts = H.create 5 in
    for n = 0 to _N - 1 do
      for i = 0 to length xTr.datav_xi.(n) - 1 do
        H.replace feature_counts xTr.datav_xi.(n).(i)
          (try H.find feature_counts xTr.datav_xi.(n).(i) + 1 with Not_found -> 1);
      done;
    done;

  let init_model = ref None in

  let run xTr' = Bfgs.optimize quiet maxi optimize_l lambda dpp _M 
                   xTr' yTr xDe yDe xTe yTe 
                   sampOption mean_model !init_model true in

  let active = H.create 5 in
  let activeCount = ref 1 in
    H.replace active 0 0;

    (match init_model0 with None -> () | Some im -> 
       for f = 0 to length im - 1 do
         let nonzero = ref false in
           for c = 0 to _C - 1 do
             if abs_float im.(f).(c) > 1e-10 then nonzero := true;
           done;
           if !nonzero then (
             H.replace active f !activeCount;
             incr activeCount;
           );
       done;
       let arr = create_matrix (ih_size active) _C 0. in
       for c = 0 to _C - 1 do
         for f = 0 to length im - 1 do
           try 
             arr.(H.find active f).(c) <- im.(f).(c);
           with Not_found -> ();
         done;
       done;
       init_model := Some arr;
    );

    let addToCurrent c y g mu =
      try
        let ct = H.find c y in
          H.replace ct g mu
      with Not_found -> (
        let ct = H.create 2 in
          H.replace ct g mu;
          H.replace c  y ct) in

    let remFromCurrent c y g =
      try
        let ct = H.find c y in
          H.remove ct g
      with Not_found -> () in
  
  let rec run_iter inum =
    let xTr' = applyActive active xTr in
    let old_active_size = ih_size active in

    let (w0,_,Some w0tx) = run xTr' in
    let priorV = ref (0. -. (lambda /. 2.) *. (Bfgs.dot w0 w0)) in
  
    let proposed  = H.create 5 in 
    let proposedN = ref 0 in
    let erroneous = H.create 5 in
  
    let baseLik   = make _N one in

      for n = 0 to _N - 1 do
        let y' = ref 0 in
        let s  = ref zero in
          for y = 0 to _C - 1 do
            if w0tx.(n).(y) > w0tx.(n).(!y') then y' := y;
            s := !s +@ w0tx.(n).(y);
          done;
          if !y' != yTr.(n) then (
            H.replace erroneous n ();
            baseLik.(n) <- w0tx.(n).(yTr.(n)) -. !s;
            for f = 0 to length xTr.datav_xi.(n) - 1 do
              if not (H.mem active xTr.datav_xi.(n).(f)) then (
                if not (mem2 proposed (yTr.(n), xTr.datav_xi.(n).(f))) then (
                  addToCurrent proposed yTr.(n) xTr.datav_xi.(n).(f) !proposedN;
                  incr proposedN;
                );
                if not (mem2 proposed (!y', xTr.datav_xi.(n).(f))) then (
                  addToCurrent proposed (!y') xTr.datav_xi.(n).(f) !proposedN;
                  incr proposedN;
                );
              );
            done;
          );
      done;

      if !proposedN > 10000 then (    (* we'll not do well with > 10k features *)
        let maxCount = ref 0 in
          iter2 (fun (_,g) _ -> maxCount := max !maxCount (H.find feature_counts g)) proposed;
        let cntCnt = make (!maxCount+1) 0 in
          iter2 (fun (_,g) _ ->
                   for i = 0 to H.find feature_counts g do
                     cntCnt.(i) <- cntCnt.(i) + 1;
                   done) proposed;
        let rec find10k i = 
          if i > !maxCount then 0
          else if cntCnt.(i) < 10000 then i
          else find10k (i+1) in
        let min_count = find10k 0 in
        let tokeep = H.create 5 in
          iter2 (fun (y,g) _ -> 
                   if H.find feature_counts g > min_count then
                     addToCurrent tokeep y g ()) proposed;
          H.clear proposed;
          proposedN := 0;
          iter2 (fun (y,g) _ -> 
                   addToCurrent proposed y g !proposedN;
                   incr proposedN) tokeep;
          Printf.fprintf stderr "Whittled down to %d features...\n" !proposedN;
      );

    let lik0 = compute_likelihood lambda xTr _C !priorV w0tx yTr erroneous None false in
  
    let szProposed = !proposedN in
    let a    = make szProposed 0. in
    let beta = make szProposed 0. in
    let pr   = make szProposed 0. in
  
      Printf.fprintf stderr "szProposed=%d" szProposed;
  
      iter2 (fun (y,g) gn ->
               let muv, av, pv = optimize_single_param baseLik lambda xTr yTr w0tx g y mu.{g+y*nf} in
                 mu.{g+y*nf} <- muv;
                 a.(gn)      <- av;
                 pr.(gn)     <- pv *@ !priorV;
                 if gn mod 100 == 0 then ( Printf.fprintf stderr "."; flush stderr );
            ) proposed;
  
      Printf.fprintf stderr "\n";
  
      Printf.fprintf stderr "beta0 = %g\n" lik0;
      flush stderr;
  
    let fin = ref false in
    let numAdded = ref 0 in
    let current = H.create 2 in
    let beta0 = ref lik0 in

    let firstIter = ref true in

      while not !fin do
        let bestBeta = ref None in
          iter2
            (fun (y,g) gn ->
               if not (H.mem active g) then (
(*                 let ll = compute_likelihood lambda xTr _C !priorV w0tx yTr erroneous (Some (y,g,mu.{g+y*nf})) false in *)
                 let ll = 
                   if !firstIter
                   then pr.(gn)
                   else compute_likelihood lambda xTr _C !priorV w0tx yTr erroneous (Some (y,g,mu.{g+y*nf})) false in
                 beta.(gn) <- 0.5 *. log (2.*.3.14159265) +. 0.5 *. log a.(gn) +. ll;
                 if is_none !bestBeta || fst (from_some !bestBeta) < beta.(gn) then
                   bestBeta := Some (beta.(gn), (y,g));
               );
            ) proposed;
          match !bestBeta with
              None -> ( Printf.fprintf stderr "--> bestBeta = None\n"; fin := true )
            | Some (betaG, (y,g)) ->
                if betaG > !beta0 then (
                  let muv = mu.{g+y*nf} in
                  Printf.fprintf stderr "Adding (y=%d,g=%s), betaG=%g, " y (get_vocab_str g) betaG; flush stderr;
                  H.replace active g !activeCount; incr activeCount;
                  addToCurrent   current  y g muv;
                  remFromCurrent proposed y g;
                  beta0 := betaG;
                  priorV := !priorV -. (lambda/.2.) *. muv *. muv;
                  ignore (compute_likelihood lambda xTr _C !priorV w0tx yTr erroneous (Some (y,g,muv)) true);
                  incr numAdded;
                  if ih_size erroneous == 0 then ( Printf.fprintf stderr "--> |erroneous| = 0\n"; fin := true );
                  if !numAdded >= _S then ( Printf.fprintf stderr "--> numAdded >= _S\n"; fin := true );
                ) else (
                  Printf.fprintf stderr "--> betaG=%g < beta0=%g\n" betaG !beta0;
                  fin := true;
                );

        firstIter := false;
      done;
  
      Printf.fprintf stderr "%d added\n" !numAdded;
  
      if !numAdded == 0 then w0 else
        let im = create_matrix (ih_size active) _C 0. in
          for c = 0 to _C - 1 do
            for f = 0 to old_active_size - 1 do
              im.(f).(c) <- w0.{f + c * old_active_size};
            done;
          done;
          iter2 (fun (y,g) mu ->
                   let gn = H.find active g in
                     im.(gn).(y) <- mu) current;
          init_model := Some im;
          run_iter (inum + 1) in

  let w_final = run_iter 1 in
  let w = make_ba _F 0. in
    H.iter (fun i j -> w.{i} <- w_final.{j}) active;
    w

  
