open Array
open Util

type modeltype = Binary | Binomial | Multiclass | Sampled

type datav =
    { datav_bernoulli : bool;  (* true if no "xv"s *)
      datav_implicit  : bool;  (* true if x^c = x^{c'} for all c,c' *)
      datav_sampled   : bool;  (* if data is for a sampling problem, this is true; assumed non-bernoulli, implicit *)
      mutable datav_numclass  : int;
      mutable datav_numfeat   : int;
      datav_indomain  : bool array;
      datav_weights   : bigarr;
      datav_xi        : int array array;
      datav_xi_c      : int array array array;
      datav_xv        : float array array;
      datav_xv_c      : float array array array }

let empty_data_set =
  { datav_bernoulli = true; datav_implicit = true; datav_numclass = -1; datav_numfeat = -1; datav_indomain = [||]; datav_weights = make_ba 0 0.; datav_xi = [||]; datav_xi_c = [||]; datav_xv = [||]; datav_xv_c = [||]; datav_sampled = false }

let dat_N x =
  if x.datav_implicit then length x.datav_xi else length x.datav_xi_c

let dat_F x =
  if x.datav_implicit then (x.datav_numfeat+1) * x.datav_numclass else x.datav_numfeat

let dat_P x n c =
  if x.datav_implicit then length x.datav_xi.(n) else length x.datav_xi_c.(n).(c)

let dat_i x n c p =
  if x.datav_implicit then x.datav_xi.(n).(p) + c * x.datav_numfeat else x.datav_xi_c.(n).(c).(p)

let dat_v x n c p =
  if x.datav_bernoulli then 1.
  else if x.datav_implicit then x.datav_xv.(n).(p) else x.datav_xv_c.(n).(c).(p)

let dat_wt x n =
  if A1.dim x.datav_weights == 0 then 1. else x.datav_weights.{n}

let dat_prep_i x n c =
  if x.datav_implicit 
  then let d = x.datav_xi.(n) in fun p -> d.(p) + c * x.datav_numfeat
  else let d = x.datav_xi_c.(n).(c) in fun p -> d.(p)

let dat_prep_v x n c =
  if x.datav_bernoulli then fun p -> 1. else
  if x.datav_implicit 
  then let d = x.datav_xv.(n) in fun p -> d.(p)
  else let d = x.datav_xv_c.(n).(c) in fun p -> d.(p)

type inputLine = EOF | EmptyLine | TestLine | DevLine | CommentLine | Success

let load_implicit quiet ber h y_read y_def sampled predict =
  let start = pos_in h in
    (* count number of examples *)
  let rec count_lines intr numtr numde numte  =
    match maybe_read_line h with
        None -> (numtr, numde, numte)
      | Some s ->
          if s = "" then count_lines intr numtr numde numte 
          else if s = "DEV"  then count_lines 1 numtr numde numte
          else if s = "TEST" then count_lines 2 numtr numde numte
          else if String.get s 0 == '#' then count_lines intr numtr numde numte
          else if intr == 0 then count_lines 0 (numtr+1) numde numte
          else if intr == 1 then count_lines 1 numtr (numde+1) numte
          else count_lines 2 numtr numde (numte+1) in
  let _ = if not quiet then (Printf.fprintf stderr "Scanning file..." ; flush stderr) in
  let (numtr, numde, numte) = count_lines 0 0 0 0 in
  let _ = seek_in h start in
  let _ = if not quiet then (Printf.fprintf stderr "%d train, %d dev, %d test, reading..." numtr numde numte ; flush stderr) in
  
    (* make arrays *)
  let tr = make numtr y_def, make numtr [||], (if ber then [||] else make numtr [||]), make_ba numtr 1., make numtr true in
  let de = make numde y_def, make numde [||], (if ber then [||] else make numde [||]), make_ba numde 1., make numde true in
  let te = make numte y_def, make numte [||], (if ber then [||] else make numte [||]), make_ba numte 1., make numte true in

  let maxFeat = ref 0 in
  let cur_line = ref 0 in

    (* our reading function *)
  let read_single_line (y,x,v,w,d) n =
    match maybe_read_line h with
        None -> EOF
      | Some s ->
          incr cur_line;

          if s = ""     then EmptyLine else
          if s = "DEV"  then DevLine   else
          if s = "TEST" then TestLine  else
          if String.get s 0 == '#' then CommentLine else
            let a = of_list (Str.split (Str.regexp "[ \t]+") s) in
            let start_idx = ref 1 in
              (* check for weight and/or ID/OOD tag *)
              if length a > 1 then (
                if a.(1) = "$$$WEIGHT" then (
                  if length a <= 2 then failwith ("$$$WEIGHT expects an argument on line " ^ string_of_int !cur_line);
                  w.{n} <- float_of_string_err !cur_line a.(2);
                  if length a > 3 && a.(3) = "$$$OUTOFDOMAIN" then (
                    d.(n) <- false;
                    start_idx := 4
                  ) else ( start_idx := 3 )
                ) else if a.(1) = "$$$OUTOFDOMAIN" then (
                  d.(n) <- false;
                  if length a > 2 && a.(2) = "$$$WEIGHT" then (
                    if length a <= 3 then failwith ("$$$WEIGHT expects an argument on line " ^ string_of_int !cur_line);
                    w.{n} <- float_of_string_err !cur_line a.(3);
                    start_idx := 4
                  ) else ( start_idx := 2 )
                )
              );

              if ber then (
                let xi = make (length a - !start_idx) 0  in
                  for p = !start_idx to length a - 1 do
                    xi.(p - !start_idx) <- get_vocab a.(p);
                    maxFeat := max !maxFeat xi.(p - !start_idx);
                  done;
                  y.(n) <- y_read !cur_line a.(0);
                  x.(n) <- xi;
                  predict y.(n) x.(n) [||];
              ) else (
                let xi = make ((length a - !start_idx)/2) 0 in
                let xv = make ((length a - !start_idx)/2) 0. in
                  for p = 0 to (length a - !start_idx) / 2 - 1 do
                    xi.(p) <- get_vocab       a.(2*p + !start_idx);
                    xv.(p) <- float_of_string_err !cur_line a.(2*p + !start_idx+1);
                    maxFeat := max !maxFeat xi.(p);
                  done;
                  y.(n) <- y_read !cur_line a.(0);
                  x.(n) <- xi;
                  v.(n) <- xv;
                  predict y.(n) x.(n) v.(n);
              );
              Success in

    (* read data *)

  let rec read_lines intr n =
    match read_single_line (if intr == 0 then tr else if intr == 1 then de else te) n with
        EOF -> ()
      | EmptyLine   -> read_lines intr  n
      | CommentLine -> read_lines intr  n
      | DevLine     -> if intr<1 then read_lines 1 0 else failwith "two 'DEV' lines"
      | TestLine    -> if intr<2 then read_lines 2 0 else failwith "two 'TEST' lines"
      | Success     -> read_lines intr (n+1) in

  let _ = read_lines 0 0 in
  let _ = if not quiet then (Printf.fprintf stderr "done\n" ; flush stderr) in
  let mk_data (y,x,v,w,d) =
    { datav_bernoulli = ber;
      datav_implicit  = true;
      datav_sampled   = sampled;
      datav_numclass  = 1;
      datav_numfeat   = !maxFeat;
      datav_indomain  = d;
      datav_weights   = w;
      datav_xi        = x;
      datav_xi_c      = [||];
      datav_xv        = if ber then [||] else v;
      datav_xv_c      = [||] }, y in

    mk_data tr, mk_data de, mk_data te




let load_explicit quiet ber h predict =
  let start = pos_in h in
    (* count number of examples *)
  let rec count_lines intr numtr numde numte  =
    match maybe_read_line h with
        None -> (numtr, numde, numte)
      | Some s ->
          if s = "" then count_lines intr numtr numde numte 
          else if s = "DEV"  then count_lines 1 numtr numde numte
          else if s = "TEST" then count_lines 2 numtr numde numte
          else if String.get s 0 == '#' then count_lines intr numtr numde numte
          else if intr == 0 then count_lines 0 (numtr+1) numde numte
          else if intr == 1 then count_lines 1 numtr (numde+1) numte
          else count_lines 2 numtr numde (numte+1) in
  let _ = if not quiet then (Printf.fprintf stderr "Scanning file..." ; flush stderr) in
  let (numtr, numde, numte) = count_lines 0 0 0 0 in
  let _ = seek_in h start in
  let _ = if not quiet then (Printf.fprintf stderr "%d train, %d dev, %d test, reading..." numtr numde numte ; flush stderr) in
  
    (* make arrays *)
  let tr = make numtr 0, make numtr [||], (if ber then [||] else make numtr [||]), make_ba numtr 1., make numtr true in
  let de = make numde 0, make numde [||], (if ber then [||] else make numde [||]), make_ba numde 1., make numde true in
  let te = make numte 0, make numte [||], (if ber then [||] else make numte [||]), make_ba numte 1., make numte true in

  let maxFeat  = ref 0 in
  let cur_line = ref 0 in
  let maxClass = ref 0 in

    (* our reading function *)
  let read_single_line (y,x,v,w,d) n =
    match maybe_read_line h with
        None -> EOF
      | Some s ->
          incr cur_line;

          if s = ""     then EmptyLine else
          if s = "DEV"  then DevLine   else
          if s = "TEST" then TestLine  else
          if String.get s 0 == '#' then CommentLine else
            let sections = of_list (Str.split (Str.regexp "[ \t]+#[ \t]+") s) in

              (* read section 0 *)
            let sec0 = of_list (Str.split (Str.regexp "[ \t]+") sections.(0)) in
              y.(n) <- int_of_string_err !cur_line sec0.(0);

              if length sec0 > 1 then (
                if sec0.(1) = "$$$WEIGHT" then (
                  if length sec0 <= 2 then failwith ("$$$WEIGHT expects an argument on line " ^ string_of_int !cur_line);
                  w.{n} <- float_of_string_err !cur_line sec0.(2);
                  if length sec0 > 3 && sec0.(3) = "$$$OUTOFDOMAIN" then d.(n) <- false;
                ) else if sec0.(1) = "$$$OUTOFDOMAIN" then (
                  d.(n) <- false;
                  if length sec0 > 2 && sec0.(2) = "$$$WEIGHT" then (
                    if length sec0 <= 3 then failwith ("$$$WEIGHT expects an argument on line " ^ string_of_int !cur_line);
                    w.{n} <- float_of_string_err !cur_line sec0.(3);
                  )
                )
              );

              maxClass := max (max !maxClass y.(n)) (length sections-1);

            let xiT = make !maxClass [||] in
            let xvT = if ber then [||] else make !maxClass [||] in

              for c = 0 to length sections - 2 do
                let a = of_list (Str.split (Str.regexp "[ \t]+") sections.(c+1)) in
                  if ber then (
                    let xi = make (length a+1) 0  in
                      xi.(0) <- get_vocab ("**BIAS**" ^ string_of_int c ^ "**");
                      maxFeat := max !maxFeat xi.(0);
                      for p = 0 to length a - 1 do
                        xi.(p+1) <- get_vocab a.(p);
                        maxFeat := max !maxFeat xi.(p+1);
                      done;
                      xiT.(c) <- xi;
                  ) else (
                    let xi = make (1+(length a)/2) 0 in
                    let xv = make (1+(length a)/2) 0. in
                      xi.(0) <- get_vocab ("**BIAS**" ^ string_of_int c ^ "**");
                      xv.(0) <- 1.;
                      maxFeat := max !maxFeat xi.(0);
                      for p = 0 to (length a) / 2 - 1 do
                        xi.(p+1) <- get_vocab                     a.(2*p);
                        xv.(p+1) <- float_of_string_err !cur_line a.(2*p+1);
                        maxFeat := max !maxFeat xi.(p+1);
                      done;
                      xiT.(c) <- xi;
                      xvT.(c) <- xv;
                  );
              done;
              x.(n) <- xiT;
              v.(n) <- xvT;

              predict y.(n) x.(n) v.(n);
              Success in

    (* read data *)

  let rec read_lines intr n =
    match read_single_line (if intr == 0 then tr else if intr == 1 then de else te) n with
        EOF -> ()
      | EmptyLine   -> read_lines intr  n
      | CommentLine -> read_lines intr  n
      | DevLine     -> if intr<1 then read_lines 1 0 else failwith "two 'DEV' lines"
      | TestLine    -> if intr<2 then read_lines 2 0 else failwith "two 'TEST' lines"
      | Success     -> read_lines intr (n+1) in

  let _ = read_lines 0 0 in
  let _ = if not quiet then (Printf.fprintf stderr "done\n" ; flush stderr) in
  let mk_data (y,x,v,w,d) =
    { datav_bernoulli = ber;
      datav_implicit  = false;
      datav_sampled   = false;
      datav_numclass  = !maxClass;
      datav_numfeat   = !maxFeat+1;
      datav_indomain  = d;
      datav_weights   = w;
      datav_xi        = [||];
      datav_xi_c      = x;
      datav_xv        = [||];
      datav_xv_c      = if ber then [||] else v }, y in

    mk_data tr, mk_data de, mk_data te

let read_twoclass_model h = 
  let ln = ref 0 in
  let w  = ref (init 32 (fun _ -> [|0.|])) in
  let rec rd () =
    match maybe_read_line h with
        None -> ()
      | Some s -> (
          incr ln;
          match Str.split (Str.regexp "[ \t]+") s with
              [feature ; value] -> (
                let f = if feature = "**BIAS**" then 0 else get_vocab feature in
                let v = float_of_string_err !ln value in
                  if f >= length !w then (  (* resize w *)
                    let w' = init (2 * f) (fun _ -> [|0.|]) in
                      for i = 0 to length !w - 1 do w'.(i).(0) <- !w.(i).(0); done;
                  w := w');
                  !w.(f).(0) <- v;
                  rd ())
            | _ -> failwith ("malformed feature/value pair on line " ^ string_of_int !ln)) in
    rd ();
    !w
                
let read_multiclass_model h = 
  let ln = ref 0 in
  let w  = ref [||] in
  let rec rd () =
    match maybe_read_line h with
        None -> ()
      | Some s -> (
          incr ln;
          match Str.split (Str.regexp "[ \t]+") s with
              (feature :: values) -> (
                let f  = if feature = "**BIAS**" then 0 else get_vocab feature in
                let vs = of_list (List.map (float_of_string_err !ln) values) in
                let _C = length vs in
                  if f >= length !w then (  (* resize w *)
                    let w' = init (2 * (f + 4)) (fun i -> if i < length !w then !w.(i) else make _C 0.) in
                      w := w');
                  iteri (fun c v -> !w.(f).(c) <- v) vs;
                  rd ())
            | _ -> failwith ("malformed feature/value pair on line " ^ string_of_int !ln)) in
    rd ();
    !w
                

let read_model h = function
    Binary | Binomial -> read_twoclass_model   h
  | Multiclass        -> read_multiclass_model h
  | Sampled           -> read_twoclass_model   h
