open Array


module BA = Bigarray
module A1 = Bigarray.Array1

type bigarr = (float, BA.float32_elt, BA.c_layout) A1.t

let copy_ba (a : bigarr) : bigarr =
  let _P = A1.dim a in
  let b  = A1.create BA.float32 BA.c_layout _P in
    for p = 0 to _P - 1 do
      b.{p} <- a.{p};
    done;
    b

let make_ba (_F : int) (v : float) : bigarr =
  let arr = A1.create BA.float32 BA.c_layout _F in
    for f = 0 to _F - 1 do arr.{f} <- v; done;
    arr

let create_matrix_ba (a : int) (b: int) (v : float) : bigarr array =
  init a (fun _ -> make_ba b v)

let fold_lefti_ba f z0 (a : bigarr) =
  let _P = A1.dim a in
  let rec fl p z =
    if p >= _P then z else
      fl (p+1) (f p a.{p} z) in
    fl 0 z0

let fold_left_ba f z0 a =
  let _P = A1.dim a in
  let rec fl p z =
    if p >= _P then z else
      fl (p+1) (f z a.{p}) in
    fl 0 z0

let iteri_ba f a =
  let _P = A1.dim a in
    for p = 0 to _P - 1 do
      f p a.{p};
    done;
    ()


let bad_float_to_zero f = match classify_float f with
    FP_normal | FP_subnormal | FP_zero -> f
  | _ -> 0.

let bad_float_to_neginf f = match classify_float f with
    FP_normal | FP_subnormal | FP_zero -> f
  | _ -> neg_infinity

let _ = Random.self_init ()

let zero  = log 0.
let one   = log 1.
let addLog (x:float) (y:float) = 
  if x == zero then y
  else if y == zero then x
  else if x -. y > 32. then x
  else if x > y then (x +. log (1. +. exp (y -. x)))
  else if y -. x > 32. then y
  else (y +. log (1. +. exp (x -. y)))
let subLog (x:float) (y:float) = 
  if y == zero then x
  else if x <= y then zero
  else if x -. y > 32. then x
  else (x +. log (1. -. exp (y -. x)))

let ( *@ ) a b = a +. b
let (+@) a b = addLog a b
let (-@) a b = subLog a b
let (^@) a b = (exp b) *. a

let (/@) a b = 
  if classify_float a == FP_infinite && classify_float b == FP_infinite
  then 
    if a < 0. && b < 0. then 0.
    else if a > 0. && b > 0. then 0.
    else a
  else a -. b

let fabs x = if x < 0. then 0. -. x else x

let fold_lefti f z0 a =
  let _P = length a in
  let rec fl p z =
    if p >= _P then z else
      fl (p+1) (f p a.(p) z) in
    fl 0 z0

let h_size h = Hashtbl.fold (fun _ _ s -> s+1) h 0
let ih_size h = IntHashtbl.fold (fun _ _ s -> s+1) h 0

let maybe_read_line h =
  try
    Some (input_line h)
  with
      End_of_file -> None

let (vocabFwd : (string,int) Hashtbl.t) = Hashtbl.create 4
let (vocabBwd : (int,string) Hashtbl.t) = Hashtbl.create 4
let vocabSize = ref 1

let get_vocab str =
  try
    Hashtbl.find vocabFwd str
  with
      Not_found -> (
        let i = !vocabSize in
          Hashtbl.replace vocabFwd str i;
          Hashtbl.replace vocabBwd i str;
          incr vocabSize;
          i
      )

let get_vocab_str i =
  if i == 0 then "**BIAS**" else
    try Hashtbl.find vocabBwd i
    with Not_found -> "**UNKNOWN**"

let printVectorVocab a =
  if length a == 0 then () else (
    if classify_float a.(0) <> FP_zero then
      Printf.printf "**BIAS** %10.20f\n" a.(0);
    for i = 1 to length a - 1 do
      try
        let s = Hashtbl.find vocabBwd i in
          if classify_float a.(i) <> FP_zero then
            Printf.printf "%s %10.20f\n" s a.(i)
      with
          Not_found -> ();
    done;
    ())

let printVectorVocab_ba a =
  if A1.dim a == 0 then () else (
    if classify_float a.{0} <> FP_zero then
      Printf.printf "**BIAS** %10.20f\n" a.{0};
    for i = 1 to A1.dim a - 1 do
      try
        let s = Hashtbl.find vocabBwd i in
          if classify_float a.{i} <> FP_zero then
            Printf.printf "%s %10.20f\n" s a.{i}
      with
          Not_found -> ();
    done;
    ())

let float_of_string_err (line_num : int) (s : string) : float =
  try float_of_string s
  with Failure err -> 
    failwith ("error: \"" ^ err ^ "\" occured on line " ^ string_of_int line_num ^ " when trying to parse \"" ^ s ^ "\" as a float")

let int_of_string_err (line_num : int) (s : string) : int =
  try int_of_string s
  with Failure err -> 
    failwith ("error: \"" ^ err ^ "\" occured on line " ^ string_of_int line_num ^ " when trying to parse \"" ^ s ^ "\" as an int")

let printIter inum maxi =
  let s = String.length (string_of_int inum) in
  let l = String.length (string_of_int maxi) in
    
    Printf.fprintf stderr "it %d%s" inum (String.make (l-s) ' ');
    flush stderr

let printDW f = Printf.fprintf stderr " dw %3.3e" f; flush stderr

let print_info quiet ppx err (tM : bigarr) tw ty (dM : bigarr) dw dy (sM : bigarr) sw sy =
  let pp = ppx tM tw ty in let de = err dM dw dy in
    if not quiet then (
      Printf.fprintf stderr " pp %5.5e er %5.5f" (0.-.pp) (err tM tw ty);
      if length dw > 0 then Printf.fprintf stderr " dpp %5.5e der %5.5f" (0.-.ppx dM dw dy) de;
      if length sw > 0 then Printf.fprintf stderr " tpp %5.5e ter %5.5f" (0.-.ppx sM sw sy) (err sM sw sy);
    );
    pp,de

let is_some = function
    Some s -> true
  | None -> false

let is_none = function
    Some s -> false
  | None -> true

let from_some = function
    Some s -> s
  | None -> failwith "from_some: None"

let bounded x = if x < 1e-6 then 1e-6 else if x > 1.-.1e-6 then 1.-.1e-6 else x

let bounded_log x = 
  if x >= -1e-50 then -1e-50
  else if x < -250. then -250. else x


let dotEx w c x v =
  let d = ref 0. in
    for i = 0 to length x - 1 do
      if x.(i) < length w then
        if c < length w.(x.(i)) then
          d := !d +. w.(x.(i)).(c) *. if length v == 0 then 1. else v.(i);
    done;
    !d
